# Copyright (C) 2026 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import ipaddress
import json
import sys
import time
import unittest

collections.Callable = collections.abc.Callable


from lib import base
from lib import utils
from lib.base import (
    BGP_FSM_ESTABLISHED,
    BGPContainer,
    assert_several_times,
    local,
    Bridge,
)
from lib.gobgp import GoBGPContainer
from lib.noseplugin import parser_option


TCP_MD5_PASSWORD = 'password'
DYNAMIC_NEIGHBOR_PREFIX = '172.17.0.0/16'
NETSHOOT_IMAGE = 'nicolaka/netshoot:v0.15'
VRF_BIND_INTERFACE = 'vrf-md5'


def _netshoot(ctn, cmd):
    return local('docker run --rm --privileged=true --net container:{0} '
                 '{1} {2}'.format(ctn.docker_name(), NETSHOOT_IMAGE, cmd),
                 capture=True)


def _expected_md5keys(ctn, peer):
    info = ctn.peers[peer]
    if info['interface'] != '':
        addr = ctn.get_neighbor(peer)['state']['neighbor_address']
    else:
        addr = info['neigh_addr'].split('/')[0]
    addr = addr.split('%')[0]
    prefixlen = 128 if ':' in addr else 32
    return 'md5keys:{0}/{1}={2}'.format(
        addr,
        prefixlen,
        TCP_MD5_PASSWORD,
    )


def _assert_md5keys(test, ctn, peer):
    expected = _expected_md5keys(ctn, peer)
    _assert_md5keys_entry(test, ctn, expected)


def _assert_md5keys_entry(test, ctn, expected, established=True,
                          listen=True):
    def f():
        if established:
            test.assertIn(expected, _netshoot(ctn, 'ss -tni'))
        if listen:
            test.assertIn(expected, _netshoot(ctn, 'ss -tlni'))

    assert_several_times(f, t=10, s=1)


def _wait_dynamic_established(ctn, peer_addr):
    def f():
        peer = json.loads(ctn.local('gobgp -j neighbor {0}'.format(peer_addr),
                                    capture=True))
        if peer['state']['session_state'] != 6:
            raise AssertionError

    assert_several_times(f, t=120, s=1)


def _setup_vrf_device(ctn, table_id):
    subnet = ipaddress.ip_network(ctn.ip_addrs[0][1], strict=False)

    # The osrg/quagga-based test image has an old iproute2 that cannot create
    # VRF devices. Run netshoot in the same network namespace and use its
    # iproute2 instead.
    _netshoot(ctn, 'ip link add {0} type vrf table {1}'.format(
        VRF_BIND_INTERFACE,
        table_id,
    ))
    _netshoot(ctn, 'ip link set dev {0} up'.format(VRF_BIND_INTERFACE))
    _netshoot(ctn, 'ip link set dev eth0 master {0}'.format(
        VRF_BIND_INTERFACE,
    ))
    _netshoot(ctn, 'ip route replace table {0} {1} dev eth0'.format(
        table_id,
        subnet,
    ))


class GoBGPTCPMD5Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='md5-g1', asn=65000,
                            router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        g2 = GoBGPContainer(name='md5-g2', asn=65001,
                            router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)

        ctns = [g1, g2]

        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        g1.add_peer(g2, passwd=TCP_MD5_PASSWORD)
        g2.add_peer(g1, passwd=TCP_MD5_PASSWORD, passive=True)

        cls.g1 = g1
        cls.g2 = g2

    def test_01_neighbor_established(self):
        self.g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g2)
        self.g2.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g1)


class GoBGPTCPMD5UnnumberedTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65000,
                            router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        g2 = GoBGPContainer(name='g2', asn=65001,
                            router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)

        ctns = [g1, g2]

        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        br = Bridge(name='unnumbered', subnet='fd00:179:5::/64')
        [br.addif(ctn) for ctn in [g1, g2]]

        utils.probe_link_local_address(g1, g2, 'eth1', 'eth1')

        g1.add_peer(g2, interface='eth1', passwd=TCP_MD5_PASSWORD)
        g2.add_peer(g1, interface='eth1', passwd=TCP_MD5_PASSWORD,
                    passive=True)

        cls.g1 = g1
        cls.g2 = g2

    def test_01_neighbor_established(self):
        self.g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g2)
        self.g2.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g1)

    def test_02_md5keys_are_set(self):
        _assert_md5keys(self, self.g1, self.g2)
        _assert_md5keys(self, self.g2, self.g1)


class GoBGPTCPMD5DynamicNeighborTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65000,
                            router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            bgp_config={
                                'peer-groups': [
                                    {
                                        'config': {
                                            'peer-group-name': 'PG1',
                                            'peer-as': 65001,
                                            'auth-password': TCP_MD5_PASSWORD,
                                        },
                                    },
                                ],
                                'dynamic-neighbors': [
                                    {
                                        'config': {
                                            'prefix': DYNAMIC_NEIGHBOR_PREFIX,
                                            'peer-group': 'PG1',
                                        },
                                    },
                                ],
                            })
        g2 = GoBGPContainer(name='g2', asn=65001,
                            router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)

        ctns = [g1, g2]

        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        g2.add_peer(g1, passwd=TCP_MD5_PASSWORD)

        cls.g1 = g1
        cls.g2 = g2
        cls.g2_addr = g2.ip_addrs[0][1].split('/')[0]

    def test_01_neighbor_established(self):
        self.g2.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g1)
        _wait_dynamic_established(self.g1, self.g2_addr)

    def test_02_md5keys_are_set(self):
        self.g2.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g1)
        _wait_dynamic_established(self.g1, self.g2_addr)
        expected = 'md5keys:{0}={1}'.format(
            DYNAMIC_NEIGHBOR_PREFIX,
            TCP_MD5_PASSWORD,
        )
        _assert_md5keys_entry(self, self.g1, expected, established=False)
        expected = 'md5keys:{0}/32={1}'.format(
            self.g2_addr,
            TCP_MD5_PASSWORD,
        )
        _assert_md5keys_entry(self, self.g1, expected, listen=False)
        _assert_md5keys(self, self.g2, self.g1)


class GoBGPTCPMD5BindInterfaceVRFTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='vrf-g1', asn=65000,
                            router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            bgp_config={
                                'global': {
                                    'config': {
                                        'bind-to-device': VRF_BIND_INTERFACE,
                                    },
                                },
                            })
        g2 = GoBGPContainer(name='vrf-g2', asn=65001,
                            router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            bgp_config={
                                'global': {
                                    'config': {
                                        'bind-to-device': VRF_BIND_INTERFACE,
                                    },
                                },
                            })

        ctns = [g1, g2]

        initial_wait_time = max(BGPContainer.run(ctn) for ctn in ctns)
        time.sleep(initial_wait_time)

        _setup_vrf_device(g1, 1010)
        _setup_vrf_device(g2, 2020)

        g1.start_gobgp()
        g2.start_gobgp()

        g1.add_peer(g2, passwd=TCP_MD5_PASSWORD, passive=True,
                    bind_interface=VRF_BIND_INTERFACE)
        g2.add_peer(g1, passwd=TCP_MD5_PASSWORD,
                    bind_interface=VRF_BIND_INTERFACE)

        cls.g1 = g1
        cls.g2 = g2

    def test_01_neighbor_established(self):
        self.g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g2)
        self.g2.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g1)

    def test_02_md5keys_are_set(self):
        # ss exposes MD5 keys through INET_DIAG_MD5SIG, but that data does not
        # include key ifindex. So we just check that the keys are set without
        # error and we don't check the ifindex part of the key.
        _assert_md5keys(self, self.g1, self.g2)
        _assert_md5keys(self, self.g2, self.g1)





# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import BGP_FSM_ESTABLISHED, local
from lib.gobgp import GoBGPContainer


class GoBGPTestBase(unittest.TestCase):

    wait_per_retry = 5
    retry_limit = 15

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        g2 = GoBGPContainer(name='g2', asn=65001, router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        g3 = GoBGPContainer(name='g3', asn=65002, router_id='192.168.0.3',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        g4 = GoBGPContainer(name='g4', asn=65003, router_id='192.168.0.4',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        ctns = [g1, g2, g3, g4]

        # advertise a route from route-server-clients
        cls.clients = {}
        for cli in [g2, g3, g4]:
            cls.clients[cli.name] = cli
        initial_wait_time = max(ctn.run() for ctn in ctns)

        time.sleep(initial_wait_time)

        for cli in cls.clients.values():
            g1.add_peer(cli, is_rs_client=True, passwd='passwd', passive=True, prefix_limit=10)
            cli.add_peer(g1, passwd='passwd')

        cls.gobgp = g1

    # test each neighbor state is turned establish
    def test_01_neighbor_established(self):
        for cli in self.clients.values():
            self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=cli)

    def test_02_softresetin_test1(self):
        g1 = self.gobgp
        g2 = self.clients['g2']
        g3 = self.clients['g3']

        p1 = {'ip-prefix': '10.0.10.0/24'}
        p2 = {'ip-prefix': '10.0.20.0/24'}

        ps0 = {'prefix-set-name': 'ps0', 'prefix-list': [p1, p2]}
        g1.set_prefix_set(ps0)

        st0 = {'conditions': {'match-prefix-set': {'prefix-set': 'ps0'}},
               'actions': {'route-disposition': 'accept-route'}}

        pol0 = {'name': 'pol0', 'statements': [st0]}

        _filename = g1.add_policy(pol0, g3, 'import', 'reject')

        g3.add_route('10.0.10.0/24')
        g3.add_route('10.0.20.0/24')

        time.sleep(1)

        num = g2.get_neighbor(g1)['state']['messages']['received'].get('update', 0)
        num = g2.get_neighbor(g1)['state']['messages']['received'].get('update', 0)

        ps0 = {'prefix-set-name': 'ps0', 'prefix-list': [p1]}
        g1.set_prefix_set(ps0)
        g1.create_config()
        # this will cause g1 to do softresetin for all neighbors (policy is changed)
        g1.reload_config()

        time.sleep(1)

        num2 = g2.get_neighbor(g1)['state']['messages']['received'].get('update', 0)
        self.assertEqual((num + 1), num2)

        g3.softreset(g1, type='out')

        time.sleep(1)

        num3 = g2.get_neighbor(g1)['state']['messages']['received'].get('update', 0)
        self.assertEqual(num2, num3)

    def test_03_softresetin_test2(self):
        g1 = self.gobgp
        g2 = self.clients['g2']

        g2.add_route('10.0.10.0/24')
        time.sleep(1)

        num = g2.get_neighbor(g1)['state']['messages']['received'].get('update', 0)
        time.sleep(3)

        g1.local('gobgp n all softresetin')
        time.sleep(3)

        num1 = g2.get_neighbor(g1)['state']['messages']['received'].get('update', 0)

        self.assertEqual(num, num1)





# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from itertools import combinations
import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import (
    BGP_FSM_IDLE,
    BGP_FSM_ESTABLISHED,
    local,
)
from lib.base import wait_for_completion
from lib.gobgp import GoBGPContainer
from lib.quagga import QuaggaBGPContainer


class GoBGPTestBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        q1 = QuaggaBGPContainer(name='q1', asn=65000, router_id='192.168.0.2')
        q2 = QuaggaBGPContainer(name='q2', asn=65000, router_id='192.168.0.3')

        qs = [q1, q2]
        ctns = [g1, q1, q2]

        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        # ibgp peer. loop topology
        for a, b in combinations(ctns, 2):
            a.add_peer(b)
            b.add_peer(a)

        # advertise a route from q1, q2
        for idx, c in enumerate(qs):
            route = '10.0.{0}.0/24'.format(idx + 1)
            c.add_route(route)

        cls.gobgp = g1
        cls.quaggas = {'q1': q1, 'q2': q2}

    # test each neighbor state is turned establish
    def test_01_neighbor_established(self):
        for q in self.quaggas.values():
            self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q)

    def test_02_check_gobgp_global_rib(self):
        for q in self.quaggas.values():
            # paths expected to exist in gobgp's global rib
            routes = list(q.routes.keys())
            timeout = 120
            interval = 1
            count = 0
            while True:
                # gobgp's global rib
                state = self.gobgp.get_neighbor_state(q)
                self.assertEqual(state, BGP_FSM_ESTABLISHED)
                global_rib = [p['prefix'] for p in self.gobgp.get_global_rib()]

                for p in global_rib:
                    if p in routes:
                        routes.remove(p)

                if len(routes) == 0:
                    break

                time.sleep(interval)
                count += interval
                if count >= timeout:
                    raise Exception('timeout')

    def test_03_check_gobgp_adj_rib_out(self):
        for q in self.quaggas.values():
            paths = self.gobgp.get_adj_rib_out(q)
            # bgp speaker mustn't forward iBGP routes to iBGP peers
            self.assertEqual(len(paths), 0)

    def test_04_originate_path(self):
        self.gobgp.add_route('10.10.0.0/24')
        dst = self.gobgp.get_global_rib('10.10.0.0/24')
        self.assertEqual(len(dst), 1)
        self.assertEqual(len(dst[0]['paths']), 1)
        path = dst[0]['paths'][0]
        self.assertEqual(path['nexthop'], '0.0.0.0')
        self.assertEqual(len(path['aspath']), 0)

    def test_05_check_gobgp_adj_rib_out(self):
        for q in self.quaggas.values():
            paths = self.gobgp.get_adj_rib_out(q)
            self.assertEqual(len(paths), len(self.gobgp.routes))
            path = paths[0]
            self.assertEqual(path['nlri']['prefix'], '10.10.0.0/24')
            peer_info = self.gobgp.peers[q]
            local_addr = peer_info['local_addr'].split('/')[0]
            self.assertEqual(path['nexthop'], local_addr)
            self.assertEqual(len(path['aspath']), 0)

    # check routes are properly advertised to all BGP speaker
    def test_06_check_quagga_global_rib(self):
        interval = 1
        timeout = int(120 / interval)
        for q in self.quaggas.values():
            done = False
            for _ in range(timeout):
                if done:
                    break
                global_rib = q.get_global_rib()
                # quagga's global_rib must have two routes at least,
                # a self-generated route and a gobgp-generated route
                if len(global_rib) < len(q.routes) + len(self.gobgp.routes):
                    time.sleep(interval)
                    continue

                peer_info = self.gobgp.peers[q]
                local_addr = peer_info['local_addr'].split('/')[0]
                for r in self.gobgp.routes:
                    self.assertTrue(r in (p['prefix'] for p in global_rib))
                    for rr in global_rib:
                        if rr['prefix'] == r:
                            self.assertEqual(rr['nexthop'], local_addr)

                for r in list(q.routes.keys()):
                    self.assertTrue(r in (p['prefix'] for p in global_rib))
                    for rr in global_rib:
                        if rr['prefix'] == r:
                            self.assertEqual(rr['nexthop'], '0.0.0.0')

                done = True
            if done:
                continue
            # should not reach here
            raise AssertionError

    def test_07_add_ebgp_peer(self):
        q3 = QuaggaBGPContainer(name='q3', asn=65001, router_id='192.168.0.4')
        self.quaggas['q3'] = q3
        initial_wait_time = q3.run()
        time.sleep(initial_wait_time)

        self.gobgp.add_peer(q3)
        q3.add_peer(self.gobgp)

        q3.add_route('10.0.3.0/24')

        self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q3)

    def test_08_check_global_rib(self):
        self.test_02_check_gobgp_global_rib()

    def test_09_check_gobgp_ebgp_adj_rib_out(self):
        q1 = self.quaggas['q1']
        q2 = self.quaggas['q2']
        q3 = self.quaggas['q3']
        paths = self.gobgp.get_adj_rib_out(q3)
        total_len = len(q1.routes) + len(q2.routes) + len(self.gobgp.routes)
        assert len(paths) == total_len
        for path in paths:
            peer_info = self.gobgp.peers[q3]
            local_addr = peer_info['local_addr'].split('/')[0]
            self.assertEqual(path['nexthop'], local_addr)
            self.assertEqual(path['aspath'], [self.gobgp.asn])

    def test_10_check_gobgp_ibgp_adj_rib_out(self):
        q1 = self.quaggas['q1']
        q3 = self.quaggas['q3']
        peer_info = self.gobgp.peers[q3]
        neigh_addr = peer_info['neigh_addr'].split('/')[0]

        for prefix in q3.routes.keys():
            paths = self.gobgp.get_adj_rib_out(q1, prefix)
            self.assertEqual(len(paths), 1)
            path = paths[0]
            # bgp router mustn't change nexthop of routes from eBGP peers
            # which are sent to iBGP peers
            self.assertEqual(path['nexthop'], neigh_addr)
            # bgp router mustn't change aspath of routes from eBGP peers
            # which are sent to iBGP peers
            self.assertEqual(path['aspath'], [q3.asn])

    # disable ebgp peer, check ebgp routes are removed
    def test_11_disable_ebgp_peer(self):
        q3 = self.quaggas['q3']
        self.gobgp.disable_peer(q3)
        del self.quaggas['q3']
        self.gobgp.wait_for(expected_state=BGP_FSM_IDLE, peer=q3)

        for route in q3.routes.keys():
            dst = self.gobgp.get_global_rib(route)
            self.assertEqual(len(dst), 0)

        for q in self.quaggas.values():
            paths = self.gobgp.get_adj_rib_out(q)
            # only gobgp's locally generated routes must exists
            print(paths)
            self.assertEqual(len(paths), len(self.gobgp.routes))

    def test_12_disable_ibgp_peer(self):
        q1 = self.quaggas['q1']
        self.gobgp.disable_peer(q1)
        self.gobgp.wait_for(expected_state=BGP_FSM_IDLE, peer=q1)

        for route in q1.routes.keys():
            dst = self.gobgp.get_global_rib(route)
            self.assertEqual(len(dst), 0)

    def test_13_enable_ibgp_peer(self):
        q1 = self.quaggas['q1']
        self.gobgp.enable_peer(q1)
        self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q1)

    def test_14_check_gobgp_adj_rib_out(self):
        for q in self.quaggas.values():
            paths = self.gobgp.get_adj_rib_out(q)
            # only gobgp's locally generated routes must exists
            self.assertEqual(len(paths), len(self.gobgp.routes))

    def test_15_add_ebgp_peer(self):
        q4 = QuaggaBGPContainer(name='q4', asn=65001, router_id='192.168.0.5')
        self.quaggas['q4'] = q4
        initial_wait_time = q4.run()
        time.sleep(initial_wait_time)

        self.gobgp.add_peer(q4)
        q4.add_peer(self.gobgp)

        prefix = '10.0.4.0/24'
        q4.add_route(prefix)

        self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q4)

        q1 = self.quaggas['q1']
        q2 = self.quaggas['q2']
        for q in [q1, q2]:
            def _f():
                return prefix in [p['nlri']['prefix'] for p in self.gobgp.get_adj_rib_out(q)]
            wait_for_completion(_f)

        def f():
            return len(q2.get_global_rib(prefix)) == 1
        wait_for_completion(f)

    def test_16_add_best_path_from_ibgp(self):
        q1 = self.quaggas['q1']
        q2 = self.quaggas['q2']

        prefix = '10.0.4.0/24'
        q1.add_route(prefix)

        def f1():
            l = self.gobgp.get_global_rib(prefix)
            return len(l) == 1 and len(l[0]['paths']) == 2
        wait_for_completion(f1)

        def f2():
            return prefix not in [p['nlri']['prefix'] for p in self.gobgp.get_adj_rib_out(q2)]
        wait_for_completion(f2)

        def f3():
            l = q2.get_global_rib(prefix)
            # route from ibgp so aspath should be empty
            return len(l) == 1 and len(l[0]['aspath']) == 0
        wait_for_completion(f3)





# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest
import inspect

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import BGP_FSM_ESTABLISHED, local
from lib.gobgp import GoBGPContainer
from lib.exabgp import ExaBGPContainer


counter = 1
_SCENARIOS = {}


def register_scenario(cls):
    global counter
    _SCENARIOS[counter] = cls
    counter += 1


def lookup_scenario(name):
    for value in list(_SCENARIOS.values()):
        if value.__name__ == name:
            return value
    return None


def wait_for(f, timeout=120):
    interval = 1
    count = 0
    while True:
        if f():
            return

        time.sleep(interval)
        count += interval
        if count >= timeout:
            raise Exception('timeout')


@register_scenario
class MalformedMpReachNlri(object):
    """
      No.1 malformaed mp-reach-nlri
    """

    @staticmethod
    def boot(env):
        gobgp_ctn_image_name = env.parser_option.gobgp_image
        log_level = env.parser_option.gobgp_log_level
        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=log_level)
        e1 = ExaBGPContainer(name='e1', asn=65001, router_id='192.168.0.2')
        e2 = ExaBGPContainer(name='e2', asn=65001, router_id='192.168.0.2')

        ctns = [g1, e1, e2]
        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        for q in [e1, e2]:
            g1.add_peer(q, is_rs_client=True)
            q.add_peer(g1)

        env.g1 = g1
        env.e1 = e1
        env.e2 = e2

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2
        for c in [e1, e2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        # advertise malformed MP_REACH_NLRI
        e1.add_route('10.7.0.17/32', attribute='0x0e 0x60 0x11223344')

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2

        # RFC 7606: gobgp treats malformed attribute flags as withdraw (no session reset)
        g1.wait_for(BGP_FSM_ESTABLISHED, e1)
        g1.wait_for(BGP_FSM_ESTABLISHED, e2)
        assert len(g1.get_global_rib()) == 0

    @staticmethod
    def executor(env):
        lookup_scenario("MalformedMpReachNlri").boot(env)
        lookup_scenario("MalformedMpReachNlri").setup(env)
        lookup_scenario("MalformedMpReachNlri").check(env)


@register_scenario
class MalformedMpUnReachNlri(object):
    """
      No.2 malformaed mp-unreach-nlri
    """

    @staticmethod
    def boot(env):
        lookup_scenario("MalformedMpReachNlri").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2
        for c in [e1, e2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        # advertise malformed MP_UNREACH_NLRI
        e1.add_route('10.7.0.17/32', attribute='0x0f 0x60 0x11223344')

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2

        # RFC 7606: gobgp treats malformed attribute flags as withdraw (no session reset)
        g1.wait_for(BGP_FSM_ESTABLISHED, e1)
        g1.wait_for(BGP_FSM_ESTABLISHED, e2)
        assert len(g1.get_global_rib()) == 0

    @staticmethod
    def executor(env):
        lookup_scenario("MalformedMpUnReachNlri").boot(env)
        lookup_scenario("MalformedMpUnReachNlri").setup(env)
        lookup_scenario("MalformedMpUnReachNlri").check(env)


@register_scenario
class MalformedAsPath(object):
    """
      No.3 malformaed as-path
    """

    @staticmethod
    def boot(env):
        lookup_scenario("MalformedMpReachNlri").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2
        for c in [e1, e2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        # advertise malformed AS_PATH
        # Send the attribute to the length and number of aspath is inconsistent
        # Attribute Type  0x02 (AS_PATH)
        # Attribute Flag  0x40 (well-known transitive)
        # Attribute Value 0x02020000ffdc (
        #  segment type    = 02
        #  segment length  = 02 -> # correct value = 01
        #  as number       = 65500   )
        e1.add_route('10.7.0.17/32', attribute='0x02 0x60 0x11223344')

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2

        # RFC 7606: gobgp treats malformed attribute flags as withdraw (no session reset)
        g1.wait_for(BGP_FSM_ESTABLISHED, e1)
        g1.wait_for(BGP_FSM_ESTABLISHED, e2)
        assert len(g1.get_global_rib()) == 0

    @staticmethod
    def executor(env):
        lookup_scenario("MalformedAsPath").boot(env)
        lookup_scenario("MalformedAsPath").setup(env)
        lookup_scenario("MalformedAsPath").check(env)


@register_scenario
class MalformedAs4Path(object):
    """
      No.4 malformaed as4-path
    """

    @staticmethod
    def boot(env):
        lookup_scenario("MalformedMpReachNlri").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2
        for c in [e1, e2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        # advertise malformed AS4_PATH
        e1.add_route('10.7.0.17/32', attribute='0x11 0x60 0x11223344')

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2

        # RFC 7606: gobgp treats malformed attribute flags as withdraw (no session reset)
        g1.wait_for(BGP_FSM_ESTABLISHED, e1)
        g1.wait_for(BGP_FSM_ESTABLISHED, e2)
        assert len(g1.get_global_rib()) == 0

    @staticmethod
    def executor(env):
        lookup_scenario("MalformedAs4Path").boot(env)
        lookup_scenario("MalformedAs4Path").setup(env)
        lookup_scenario("MalformedAs4Path").check(env)


@register_scenario
class MalformedNexthop(object):
    """
      No.5 malformaed nexthop
    """

    @staticmethod
    def boot(env):
        lookup_scenario("MalformedMpReachNlri").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2
        for c in [e1, e2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        # advertise malformed NEXT_HOP
        # 0x0e: MP_REACH_NLRI
        # 0x60: Optional, Transitive
        # 0x01: AFI(IPv4)
        # 0x01: SAFI(unicast)
        # 0x10: Length of Next Hop Address
        # 0xffffff00: Network address of Next Hop
        # 0x00: Reserved
        e1.add_route('10.7.0.17/32', attribute='0x0e 0x60 0x010110ffffff0000')

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2

        # RFC 7606: gobgp treats malformed attribute flags as withdraw (no session reset)
        g1.wait_for(BGP_FSM_ESTABLISHED, e1)
        g1.wait_for(BGP_FSM_ESTABLISHED, e2)
        assert len(g1.get_global_rib()) == 0

    @staticmethod
    def executor(env):
        lookup_scenario("MalformedNexthop").boot(env)
        lookup_scenario("MalformedNexthop").setup(env)
        lookup_scenario("MalformedNexthop").check(env)


@register_scenario
class MalformedRouteFamily(object):
    """
      No.6 malformaed route family
    """

    @staticmethod
    def boot(env):
        lookup_scenario("MalformedMpReachNlri").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2
        for c in [e1, e2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        # advertise malformed ROUTE_FAMILY
        # 0x0e: MP_REACH_NLRI
        # 0x60: Optional, Transitive
        # 0x01: AFI(IPv4)
        # 0x01: SAFI(unicast)
        # 0x10: Length of Next Hop Address
        # 0xffffff00: Network address of Next Hop
        # 0x00: Reserved
        e1.add_route('10.7.0.17/32', attribute='0x0e 0x60 0x0002011020010db800000000000000000000000100')

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2

        # RFC 7606: gobgp treats malformed attribute flags as withdraw (no session reset)
        g1.wait_for(BGP_FSM_ESTABLISHED, e1)
        g1.wait_for(BGP_FSM_ESTABLISHED, e2)
        assert len(g1.get_global_rib()) == 0

    @staticmethod
    def executor(env):
        lookup_scenario("MalformedRouteFamily").boot(env)
        lookup_scenario("MalformedRouteFamily").setup(env)
        lookup_scenario("MalformedRouteFamily").check(env)


@register_scenario
class MalformedAsPathSegmentLengthInvalid(object):
    """
      No.7 malformaed aspath segment length invalid
    """

    @staticmethod
    def boot(env):
        lookup_scenario("MalformedMpReachNlri").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2
        for c in [e1, e2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        # advertise malformed AS_PATH SEGMENT LENGTH
        # Send the attribute to the length and number of aspath is inconsistent
        # Attribute Type  0x02 (AS_PATH)
        # Attribute Flag  0x40 (well-known transitive)
        # Attribute Value 0x02020000ffdc (
        #  segment type    = 02
        #  segment length  = 02 -> # correct value = 01
        #  as number       = 65500   )
        e1.add_route('10.7.0.17/32', attribute='0x02 0x40 0x0202ffdc')

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2

        # RFC 7606: gobgp treats malformed AS_PATH as withdraw (no session reset)
        g1.wait_for(BGP_FSM_ESTABLISHED, e1)
        g1.wait_for(BGP_FSM_ESTABLISHED, e2)
        assert len(g1.get_global_rib()) == 0

    @staticmethod
    def executor(env):
        lookup_scenario("MalformedAsPathSegmentLengthInvalid").boot(env)
        lookup_scenario("MalformedAsPathSegmentLengthInvalid").setup(env)
        lookup_scenario("MalformedAsPathSegmentLengthInvalid").check(env)


@register_scenario
class MalformedNexthopLoopbackAddr(object):
    """
      No.8 malformaed nexthop loopback addr
    """

    @staticmethod
    def boot(env):
        lookup_scenario("MalformedMpReachNlri").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2
        for c in [e1, e2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        # Malformed Invalid NEXT_HOP Attribute
        # Send the attribute of invalid nexthop
        # next-hop 127.0.0.1 -> # correct value = other than loopback and 0.0.0.0 address
        e1.add_route('10.7.0.17/32', nexthop='127.0.0.1')

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2

        # RFC 7606: gobgp treats invalid NEXT_HOP as withdraw (no session reset)
        g1.wait_for(BGP_FSM_ESTABLISHED, e1)
        g1.wait_for(BGP_FSM_ESTABLISHED, e2)
        assert len(g1.get_global_rib()) == 0

    @staticmethod
    def executor(env):
        lookup_scenario("MalformedNexthopLoopbackAddr").boot(env)
        lookup_scenario("MalformedNexthopLoopbackAddr").setup(env)
        lookup_scenario("MalformedNexthopLoopbackAddr").check(env)


@register_scenario
class MalformedOriginType(object):
    """
      No.9 malformaed origin type
    """

    @staticmethod
    def boot(env):
        lookup_scenario("MalformedMpReachNlri").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2
        for c in [e1, e2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        # Invalid ORIGIN Attribute
        # Send the attribute of origin type 4
        # Attribute Type  0x01 (Origin)
        # Attribute Flag  0x40 (well-known transitive)
        # Attribute Value 0x04 (
        #  origin type    = 04 -> # correct value = 01 or 02 or 03 )
        e1.add_route('10.7.0.17/32', attribute='0x1 0x40 0x04')

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        e2 = env.e2

        # RFC 7606: gobgp treats invalid ORIGIN as withdraw (no session reset)
        g1.wait_for(BGP_FSM_ESTABLISHED, e1)
        g1.wait_for(BGP_FSM_ESTABLISHED, e2)
        assert len(g1.get_global_rib()) == 0

    @staticmethod
    def executor(env):
        lookup_scenario("MalformedOriginType").boot(env)
        lookup_scenario("MalformedOriginType").setup(env)
        lookup_scenario("MalformedOriginType").check(env)


class TestGoBGPBase(unittest.TestCase):

    wait_per_retry = 5
    retry_limit = 10

    @classmethod
    def setUpClass(cls):
        idx = parser_option.test_index
        base.TEST_PREFIX = parser_option.test_prefix
        cls.parser_option = parser_option
        cls.executors = []
        if idx == 0:
            print('unset test-index. run all test sequential')
            for _, v in list(_SCENARIOS.items()):
                for k, m in inspect.getmembers(v, inspect.isfunction):
                    if k == 'executor':
                        cls.executor = m
                cls.executors.append(cls.executor)
        elif idx not in _SCENARIOS:
            print('invalid test-index. # of scenarios: {0}'.format(len(_SCENARIOS)))
            sys.exit(1)
        else:
            for k, m in inspect.getmembers(_SCENARIOS[idx], inspect.isfunction):
                if k == 'executor':
                    cls.executor = m
            cls.executors.append(cls.executor)

    def test(self):
        for e in self.executors:
            e(self)
            print('[PASS] %s' % e.__qualname__.split('.')[0], file=sys.stderr, flush=True)





# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import BGP_FSM_ESTABLISHED, local
from lib.gobgp import GoBGPContainer
from lib.exabgp import ExaBGPContainer
from lib.yabgp import YABGPContainer


class FlowSpecTest(unittest.TestCase):

    """
    Test case for Flow Specification.
    """
    # +------------+            +------------+
    # | G1(GoBGP)  |---(iBGP)---| E1(ExaBGP) |
    # | 172.17.0.2 |            | 172.17.0.3 |
    # +------------+            +------------+
    #      |
    #    (iBGP)
    #      |
    # +------------+
    # | Y1(YABGP)  |
    # | 172.17.0.4 |
    # +------------+

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        cls.g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                                ctn_image_name=gobgp_ctn_image_name,
                                log_level=parser_option.gobgp_log_level)

        cls.e1 = ExaBGPContainer(name='e1', asn=65000, router_id='192.168.0.2')

        cls.y1 = YABGPContainer(name='y1', asn=65000, router_id='192.168.0.3')

        ctns = [cls.g1, cls.e1, cls.y1]
        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        # Add FlowSpec routes into GoBGP.
        cls.g1.add_route(
            route='ipv4/all',
            rf='ipv4-flowspec',
            matchs=[
                'destination 11.1.0.0/24',
                'source 11.2.0.0/24',
                "protocol '==tcp &=udp icmp >igmp >=egp <ipip <=rsvp !=gre'",
                "port '==80 &=90 8080 >9090 >=8180 <9190 <=8081 !=9091 &!443'",
                'destination-port 80',
                'source-port 8080',
                'icmp-type 0',
                'icmp-code 2',
                "tcp-flags '==S &=SA A !F !=U =!R'",
                'packet-length 100',
                'dscp 12',
                'fragment dont-fragment is-fragment+first-fragment',
            ],
            thens=['discard'])
        cls.g1.add_route(
            route='ipv6/dst/src/label',  # others are tested on IPv4
            rf='ipv6-flowspec',
            matchs=[
                'destination 2001:1::/64 10',
                'source 2001:2::/64 15',
                'label 12',
            ],
            thens=['discard'])

        cls.g1.add_peer(cls.e1, flowspec=True)
        cls.e1.add_peer(cls.g1, flowspec=True)
        cls.g1.add_peer(cls.y1, flowspec=True)
        cls.y1.add_peer(cls.g1, flowspec=True)

        cls.g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=cls.e1)
        cls.g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=cls.y1)

        # Add FlowSpec routes into ExaBGP.
        cls.e1.add_route(
            route='ipv4/all',
            rf='ipv4-flowspec',
            matchs=[
                'destination 12.1.0.0/24',
                'source 12.2.0.0/24',
                'protocol =tcp',
                'port >=80',
                'destination-port >5000',
                'source-port 8080',
                'icmp-type <1',
                'icmp-code <=2',
                "tcp-flags FIN",
                'packet-length >100&<200',
                'dscp 12',
                'fragment dont-fragment',
            ],
            thens=['discard'])
        cls.e1.add_route(
            route='ipv6/dst/src/protocol/label',  # others are tested on IPv4
            rf='ipv6-flowspec',
            matchs=[
                'destination 2002:1::/64/10',
                'source 2002:2::/64/15',
                'next-header udp',
                'flow-label >100',
            ],
            thens=['discard'])

        # Add FlowSpec routes into YABGP.
        cls.y1.add_route(
            route='ipv4/all',
            rf='ipv4-flowspec',
            matchs=[
                'destination 13.1.0.0/24',
                'source 13.2.0.0/24',
                "protocol =6|=17",
                # "port",  # not seem to be supported
                'destination-port =80',
                'source-port <8080|>9090',
                'icmp-type >=1',
                'icmp-code <2',
                # "tcp-flags",  # not seem to be supported via REST API
                'packet-length <=100',
                'dscp =12',
                # "fragment",  # not seem to be supported via REST API
            ],
            thens=['traffic-rate:0:0'])  # 'discard'
        # IPv6 FlowSpec: not supported with YABGP v0.4.0
        # cls.y1.add_route(
        #     route='ipv6/dst/src/label',  # others are tested on IPv4
        #     rf='ipv6-flowspec',
        #     matchs=[
        #         'destination 2003:1::/64/10',
        #         'source 2003:2::/64/15',
        #         'label 12',
        #     ],
        #     thens=['traffic-rate:0:0'])  # 'discard'

    def test_01_ipv4_exabgp_adj_rib_in(self):
        rib = self.e1.get_adj_rib_in(self.g1, rf='ipv4-flowspec')
        self.assertEqual(1, len(rib))
        nlri = list(rib)[0]  # advertised from G1(GoBGP)
        _exp_fmt = (
            # INPUTS:
            # 'destination 11.1.0.0/24',
            "destination-ipv4 11.1.0.0/24 "
            # 'source 11.2.0.0/24',
            "source-ipv4 11.2.0.0/24 "
            # "protocol '==tcp &=udp icmp >igmp >=egp <ipip <=rsvp !=gre'",
            "protocol [ =tcp&=udp =icmp >igmp >=egp <ipip <=rsvp !=gre ] "
            # "port '==80 &=90 8080 >9090 >=8180 <9190 <=8081 !=9091 &!443'",
            "port [ =80&=90 =8080 >9090 >=8180 <9190 <=8081 !=9091&!=443 ] "
            # 'destination-port 80',
            "destination-port =80 "
            # 'source-port 8080',
            "source-port =8080 "
            # 'icmp-type 0',
            "icmp-type =echo-reply "
            # 'icmp-code 2',
            "icmp-code =2 "
            # "tcp-flags '==S &=SA A !F !=U =!R'",
            "tcp-flags [ =syn&=%s ack !fin !=urgent !=rst ] "
            # 'packet-length 100',
            "packet-length =100 "
            # 'dscp 12',
            "dscp =12 "
            # 'fragment dont-fragment is-fragment+first-fragment',
            "fragment [ dont-fragment is-fragment+first-fragment ]"
        )
        # Note: Considers variants of SYN + ACK
        expected_list = (_exp_fmt % 'syn+ack', _exp_fmt % 'ack+syn')
        self.assertIn(nlri, expected_list)

    def test_02_ipv6_exabgp_adj_rib_in(self):
        rib = self.e1.get_adj_rib_in(self.g1, rf='ipv6-flowspec')
        self.assertEqual(1, len(rib))
        nlri = list(rib)[0]  # advertised from G1(GoBGP)
        expected = (
            # INPUTS:
            # 'destination 2001:1::/64 10',
            "destination-ipv6 2001:1::/64/10 "
            # 'source 2001:2::/64 15',
            "source-ipv6 2001:2::/64/15 "
            # 'label 12',
            "flow-label =12"
        )
        self.assertEqual(expected, nlri)

    def test_03_ipv4_yabgp_adj_rib_in(self):
        rib = self.y1.get_adj_rib_in(peer=self.g1, rf='flowspec')
        self.assertEqual(1, len(rib))
        nlri = list(rib)[0]  # advertised from G1(GoBGP)
        expected = (
            # INPUTS:
            # 'destination 11.1.0.0/24',
            '{"1": "11.1.0.0/24",'
            # 'source 11.2.0.0/24',
            ' "2": "11.2.0.0/24",'
            # "protocol '==tcp &=udp icmp >igmp >=egp <ipip <=rsvp !=gre'",
            ' "3": "=6&=17|=1|>2|>=8|<94|<=46|><47",'
            # "port '==80 &=90 8080 >9090 >=8180 <9190 <=8081 !=9091 &!443'",
            ' "4": "=80&=90|=8080|>9090|>=8180|<9190|<=8081|><9091&><443",'
            # 'destination-port 80',
            ' "5": "=80",'
            # 'source-port 8080',
            ' "6": "=8080",'
            # 'icmp-type 0',
            ' "7": "=0",'
            # 'icmp-code 2',
            ' "8": "=2",'
            # "tcp-flags '==S &=SA A !F !=U =!R'",
            ' "9": "=2&=18|16|>1|>=32|>=4",'
            # 'packet-length 100',
            ' "10": "=100",'
            # 'dscp 12',
            ' "11": "=12",'
            # 'fragment dont-fragment is-fragment+first-fragment',
            ' "12": "1|6"}'
        )
        self.assertEqual(expected, nlri)

    def test_04_ipv6_yabgp_adj_rib_in(self):
        # IPv6 FlowSpec: not supported with YABGP v0.4.0
        pass

    def test_05_ipv4_gobgp_global_rib(self):
        rib = self.g1.get_global_rib(rf='ipv4-flowspec')
        self.assertEqual(3, len(rib))
        output_nlri_list = [r['prefix'] for r in rib]
        nlri_g1 = (
            # INPUTS:
            # 'destination 11.1.0.0/24',
            "[destination: 11.1.0.0/24]"
            # 'source 11.2.0.0/24',
            "[source: 11.2.0.0/24]"
            # "protocol '==tcp &=udp icmp >igmp >=egp <ipip <=rsvp !=gre'",
            "[protocol: ==tcp&==udp ==icmp >igmp >=egp <ipip <=rsvp !=gre]"
            # "port '==80 &=90 8080 >9090 >=8180 <9190 <=8081 !=9091 &!443'",
            "[port: ==80&==90 ==8080 >9090 >=8180 <9190 <=8081 !=9091&!=443]"
            # 'destination-port 80',
            "[destination-port: ==80]"
            # 'source-port 8080',
            "[source-port: ==8080]"
            # 'icmp-type 0',
            "[icmp-type: ==0]"
            # 'icmp-code 2',
            "[icmp-code: ==2]"
            # "tcp-flags '==S &=SA A !F !=U =!R'",
            "[tcp-flags: =S&=SA A !F !=U !=R]"
            # 'packet-length 100',
            "[packet-length: ==100]"
            # 'dscp 12',
            "[dscp: ==12]"
            # 'fragment dont-fragment is-fragment+first-fragment',
            "[fragment: dont-fragment is-fragment+first-fragment]"
        )
        nlri_e1 = (
            # INPUTS:
            # 'destination 12.1.0.0/24',
            '[destination: 12.1.0.0/24]'
            # 'source 12.2.0.0/24',
            '[source: 12.2.0.0/24]'
            # 'protocol =tcp',
            '[protocol: ==tcp]'
            # 'port >=80',
            '[port: >=80]'
            # 'destination-port >5000',
            '[destination-port: >5000]'
            # 'source-port 8080',
            '[source-port: ==8080]'
            # 'icmp-type <1',
            '[icmp-type: <1]'
            # 'icmp-code <=2',
            '[icmp-code: <=2]'
            # "tcp-flags FIN",
            '[tcp-flags: F]'
            # 'packet-length >100&<200',
            '[packet-length: >100&<200]'
            # 'dscp 12',
            '[dscp: ==12]'
            # 'fragment dont-fragment',
            '[fragment: dont-fragment]'
        )
        nlri_y1 = (
            # INPUTS:
            # 'destination 13.1.0.0/24',
            "[destination: 13.1.0.0/24]"
            # 'source 13.2.0.0/24',
            "[source: 13.2.0.0/24]"
            # "protocol =6|=17",
            "[protocol: ==tcp ==udp]"
            # 'destination-port =80',
            "[destination-port: ==80]"
            # 'source-port <8080|>9090',
            "[source-port: <8080 >9090]"
            # 'icmp-type >=1',
            "[icmp-type: >=1]"
            # 'icmp-code <2',
            "[icmp-code: <2]"
            # 'packet-length <=100',
            "[packet-length: <=100]"
            # 'dscp =12',
            "[dscp: ==12]"
        )
        for nlri in [nlri_g1, nlri_e1, nlri_y1]:
            self.assertIn(nlri, output_nlri_list)

    def test_06_ipv6_gobgp_global_rib(self):
        rib = self.g1.get_global_rib(rf='ipv6-flowspec')
        import json
        self.assertEqual(2, len(rib), json.dumps(rib))
        output_nlri_list = [r['prefix'] for r in rib]
        nlri_g1 = (
            # INPUTS:
            # 'destination 2001:1::/64 10',
            "[destination: 2001:1::/64/10]"
            # 'source 2001:2::/64 15',
            "[source: 2001:2::/64/15]"
            # 'label 12',
            "[label: ==12]"
        )
        nlri_e1 = (
            # INPUTS:
            # 'destination 2002:1::/64/10',
            '[destination: 2002:1::/64/10]'
            # 'source 2002:2::/64/15',
            '[source: 2002:2::/64/15]'
            # 'next-header udp',
            '[protocol: ==udp]'
            # 'flow-label >100',
            '[label: >100]'
        )
        for nlri in [nlri_g1, nlri_e1]:
            self.assertIn(nlri, output_nlri_list)

    def test_07_ipv4_exabgp_delete_route(self):
        # Delete a route on E1(ExaBGP)
        self.e1.del_route(route='ipv4/all')
        time.sleep(1)
        # Test if the route is deleted or not
        rib = self.g1.get_adj_rib_in(peer=self.e1, rf='ipv4-flowspec')
        self.assertEqual(0, len(rib))

    def test_08_ipv6_exabgp_delete_route(self):
        # Delete a route on E1(ExaBGP)
        self.e1.del_route(route='ipv6/dst/src/protocol/label')
        time.sleep(1)
        # Test if the route is deleted or not
        rib = self.g1.get_adj_rib_in(peer=self.e1, rf='ipv6-flowspec')
        self.assertEqual(0, len(rib))

    def test_09_ipv4_yabgp_delete_route(self):
        # Delete a route on Y1(YABGP)
        self.y1.del_route(route='ipv4/all')
        time.sleep(1)
        # Test if the route is deleted or not
        rib = self.g1.get_adj_rib_in(peer=self.y1, rf='ipv4-flowspec')
        self.assertEqual(0, len(rib))

    def test_10_ipv6_yabgp_delete_route(self):
        # IPv6 FlowSpec: not supported with YABGP v0.4.0
        pass

    def test_11_ipv4_gobgp_delete_route(self):
        # Delete a route on G1(GoBGP)
        self.g1.del_route(route='ipv4/all')
        time.sleep(1)
        # Test if the route is deleted or not
        rib_e1 = self.e1.get_adj_rib_in(peer=self.g1, rf='ipv4-flowspec')
        self.assertEqual(0, len(rib_e1))
        rib_y1 = self.y1.get_adj_rib_in(peer=self.g1, rf='ipv4-flowspec')
        self.assertEqual(0, len(rib_y1))

    def test_12_ipv6_gobgp_delete_route(self):
        # Delete a route on G1(GoBGP)
        self.g1.del_route(route='ipv6/dst/src/label')
        time.sleep(1)
        # Test if the route is deleted or not
        rib_e1 = self.e1.get_adj_rib_in(peer=self.g1, rf='ipv6-flowspec')
        self.assertEqual(0, len(rib_e1))
        rib_y1 = self.y1.get_adj_rib_in(peer=self.g1, rf='ipv6-flowspec')
        self.assertEqual(0, len(rib_y1))





#!/usr/bin/env python

"""
DDoS auto-mitigation via FlowSpec using GoBGP's gRPC API.

Runs a lightweight HTTP server that receives attack alerts from a DDoS
detection system and dynamically adds or removes FlowSpec rules through
GoBGP. When an attack is detected, the corresponding FlowSpec rule is
announced to upstream routers. When the attack clears, the rule is
withdrawn.

The webhook interface is generic and works with any system that can
POST JSON (custom sFlow analyzers, monitoring tools, etc.).

Usage:
    # Generate gRPC bindings first (see docs/sources/grpc-client.md)
    $ cd tools/grpc/python
    $ PYTHONPATH=$PYTHONPATH:./api python3 flowspec_mitigation.py

    # Add a FlowSpec rule to drop UDP traffic to 203.0.113.1/32:
    $ curl -X POST http://localhost:8080/mitigate -H 'Content-Type: application/json' -d '{
        "destination": "203.0.113.1/32",
        "protocol": 17,
        "action": "discard"
      }'

    # Add a rate-limit rule for TCP port 80 traffic:
    $ curl -X POST http://localhost:8080/mitigate -H 'Content-Type: application/json' -d '{
        "destination": "203.0.113.0/24",
        "protocol": 6,
        "destination_port": 80,
        "action": "rate-limit",
        "rate": 1000.0
      }'

    # Remove a rule by its UUID (returned from the mitigate response):
    $ curl -X POST http://localhost:8080/clear -H 'Content-Type: application/json' -d '{
        "uuid": "hex-uuid-from-mitigate-response"
      }'

    # List active FlowSpec rules:
    $ curl http://localhost:8080/rules
"""

from __future__ import absolute_import
from __future__ import print_function

import json
import sys
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer

import grpc

import attribute_pb2
import common_pb2
import extcom_pb2
import gobgp_pb2
import gobgp_pb2_grpc
import nlri_pb2

_TIMEOUT_SECONDS = 10
_GOBGP_GRPC_ADDR = "localhost:50051"
_LISTEN_ADDR = "0.0.0.0"
_LISTEN_PORT = 8080

# RFC 5575 numeric operator flags (see also gobgp constant.go)
_OP_EQ = 0x01
_OP_END = 0x80

# FlowSpec component types (RFC 5575 Section 4)
_TYPE_DST_PREFIX = 1
_TYPE_SRC_PREFIX = 2
_TYPE_PROTOCOL = 3
_TYPE_DST_PORT = 5
_TYPE_SRC_PORT = 6

# Common IP protocol numbers
_PROTO_ICMP = 1
_PROTO_TCP = 6
_PROTO_UDP = 17
_PROTO_GRE = 47

# Active rules keyed by UUID for clean withdrawal
_active_rules = {}
_lock = threading.Lock()


def _build_flowspec_rules(params):
    """Build FlowSpec NLRI rules from request parameters."""
    rules = []

    if "destination" in params:
        prefix, prefix_len = params["destination"].rsplit("/", 1)
        rules.append(nlri_pb2.FlowSpecRule(
            ip_prefix=nlri_pb2.FlowSpecIPPrefix(
                type=_TYPE_DST_PREFIX,
                prefix_len=int(prefix_len),
                prefix=prefix,
            )
        ))

    if "source" in params:
        prefix, prefix_len = params["source"].rsplit("/", 1)
        rules.append(nlri_pb2.FlowSpecRule(
            ip_prefix=nlri_pb2.FlowSpecIPPrefix(
                type=_TYPE_SRC_PREFIX,
                prefix_len=int(prefix_len),
                prefix=prefix,
            )
        ))

    if "protocol" in params:
        rules.append(nlri_pb2.FlowSpecRule(
            component=nlri_pb2.FlowSpecComponent(
                type=_TYPE_PROTOCOL,
                items=[nlri_pb2.FlowSpecComponentItem(
                    op=_OP_END | _OP_EQ,
                    value=int(params["protocol"]),
                )],
            )
        ))

    if "destination_port" in params:
        rules.append(nlri_pb2.FlowSpecRule(
            component=nlri_pb2.FlowSpecComponent(
                type=_TYPE_DST_PORT,
                items=[nlri_pb2.FlowSpecComponentItem(
                    op=_OP_END | _OP_EQ,
                    value=int(params["destination_port"]),
                )],
            )
        ))

    if "source_port" in params:
        rules.append(nlri_pb2.FlowSpecRule(
            component=nlri_pb2.FlowSpecComponent(
                type=_TYPE_SRC_PORT,
                items=[nlri_pb2.FlowSpecComponentItem(
                    op=_OP_END | _OP_EQ,
                    value=int(params["source_port"]),
                )],
            )
        ))

    return rules


def _build_action(params):
    """Build the FlowSpec extended community action."""
    action = params.get("action", "discard")

    if action == "discard":
        return extcom_pb2.ExtendedCommunity(
            traffic_rate=extcom_pb2.TrafficRateExtended(rate=0.0)
        )
    elif action == "rate-limit":
        rate = float(params.get("rate", 0.0))
        return extcom_pb2.ExtendedCommunity(
            traffic_rate=extcom_pb2.TrafficRateExtended(rate=rate)
        )
    elif action == "redirect":
        asn = int(params.get("redirect_asn", 0))
        local_admin = int(params.get("redirect_local_admin", 0))
        return extcom_pb2.ExtendedCommunity(
            redirect_two_octet_as_specific=(
                extcom_pb2.RedirectTwoOctetAsSpecificExtended(
                    asn=asn,
                    local_admin=local_admin,
                )
            )
        )
    else:
        # Default to accept (no extended community needed)
        return None


def _detect_family(params):
    """Detect whether to use IPv4 or IPv6 FlowSpec based on prefixes."""
    for key in ("destination", "source"):
        if key in params and ":" in params[key].split("/")[0]:
            return common_pb2.Family(
                afi=common_pb2.Family.AFI_IP6,
                safi=common_pb2.Family.SAFI_FLOW_SPEC_UNICAST,
            )
    return common_pb2.Family(
        afi=common_pb2.Family.AFI_IP,
        safi=common_pb2.Family.SAFI_FLOW_SPEC_UNICAST,
    )


def add_flowspec_rule(stub, params):
    """Add a FlowSpec rule to GoBGP and return its UUID."""
    rules = _build_flowspec_rules(params)
    if not rules:
        raise ValueError("no match criteria specified")

    nlri = nlri_pb2.NLRI(
        flow_spec=nlri_pb2.FlowSpecNLRI(rules=rules)
    )

    attributes = [
        attribute_pb2.Attribute(
            origin=attribute_pb2.OriginAttribute(origin=0)
        ),
    ]

    action_community = _build_action(params)
    if action_community is not None:
        attributes.append(
            attribute_pb2.Attribute(
                extended_communities=(
                    attribute_pb2.ExtendedCommunitiesAttribute(
                        communities=[action_community]
                    )
                )
            )
        )

    family = _detect_family(params)

    response = stub.AddPath(
        gobgp_pb2.AddPathRequest(
            table_type=gobgp_pb2.TABLE_TYPE_GLOBAL,
            path=gobgp_pb2.Path(
                nlri=nlri,
                pattrs=attributes,
                family=family,
            ),
        ),
        timeout=_TIMEOUT_SECONDS,
    )

    uuid = response.uuid.hex()
    with _lock:
        _active_rules[uuid] = params
    return uuid


def delete_flowspec_rule(stub, uuid):
    """Remove a FlowSpec rule by its UUID."""
    stub.DeletePath(
        gobgp_pb2.DeletePathRequest(
            table_type=gobgp_pb2.TABLE_TYPE_GLOBAL,
            uuid=bytes.fromhex(uuid),
        ),
        timeout=_TIMEOUT_SECONDS,
    )

    with _lock:
        _active_rules.pop(uuid, None)


def list_flowspec_rules(stub):
    """List active FlowSpec routes from the global RIB."""
    result = []
    for family_afi in (common_pb2.Family.AFI_IP, common_pb2.Family.AFI_IP6):
        family = common_pb2.Family(
            afi=family_afi,
            safi=common_pb2.Family.SAFI_FLOW_SPEC_UNICAST,
        )
        try:
            paths = stub.ListPath(
                gobgp_pb2.ListPathRequest(
                    table_type=gobgp_pb2.TABLE_TYPE_GLOBAL,
                    family=family,
                ),
                timeout=_TIMEOUT_SECONDS,
            )
            for dest in paths:
                for path in dest.destination.paths:
                    result.append(path.nlri.flow_spec)
        except grpc.RpcError:
            pass
    return result


class MitigationHandler(BaseHTTPRequestHandler):
    stub = None

    def _send_json(self, code, data):
        body = json.dumps(data).encode()
        self.send_response(code)
        self.send_header("Content-Type", "application/json")
        self.send_header("Content-Length", str(len(body)))
        self.end_headers()
        self.wfile.write(body)

    def _read_body(self):
        length = int(self.headers.get("Content-Length", 0))
        return json.loads(self.rfile.read(length)) if length else {}

    def do_POST(self):
        if self.path == "/mitigate":
            try:
                params = self._read_body()
                uuid = add_flowspec_rule(self.stub, params)
                print("[+] added rule %s: %s" % (uuid[:12], params))
                self._send_json(200, {"status": "ok", "uuid": uuid})
            except Exception as e:
                print("[-] error adding rule: %s" % e)
                self._send_json(400, {"status": "error", "message": str(e)})

        elif self.path == "/clear":
            try:
                params = self._read_body()
                uuid = params.get("uuid", "")
                delete_flowspec_rule(self.stub, uuid)
                print("[+] removed rule %s" % uuid[:12])
                self._send_json(200, {"status": "ok"})
            except Exception as e:
                print("[-] error removing rule: %s" % e)
                self._send_json(400, {"status": "error", "message": str(e)})

        else:
            self._send_json(404, {"status": "not found"})

    def do_GET(self):
        if self.path == "/rules":
            with _lock:
                rules = dict(_active_rules)
            self._send_json(200, {"rules": rules})
        else:
            self._send_json(404, {"status": "not found"})

    def log_message(self, format, *args):
        pass  # Suppress default access logs


def run():
    channel = grpc.insecure_channel(_GOBGP_GRPC_ADDR)
    stub = gobgp_pb2_grpc.GoBgpServiceStub(channel)
    MitigationHandler.stub = stub

    server = HTTPServer((_LISTEN_ADDR, _LISTEN_PORT), MitigationHandler)
    print("FlowSpec mitigation webhook listening on %s:%d" % (
        _LISTEN_ADDR, _LISTEN_PORT))
    print("GoBGP gRPC target: %s" % _GOBGP_GRPC_ADDR)
    print("")
    print("Endpoints:")
    print("  POST /mitigate  - add a FlowSpec rule")
    print("  POST /clear     - remove a FlowSpec rule by UUID")
    print("  GET  /rules     - list active rules")
    print("")
    print("Verify with: gobgp global rib -a ipv4-flowspec")

    try:
        server.serve_forever()
    except KeyboardInterrupt:
        print("\nShutting down")
        server.server_close()


if __name__ == "__main__":
    run()
